module MetalAdventures
  class Dice
    class Modifiers
      attr_accessor :reduced, :list, :priority

      # normal modifier = all dice
      # partial modifier = only one dice
      # 2 partial change to one onrmal
      # sc and pmf cancel e2f
      #
      # :sc|:SC   : "seconde chance", reroll failed dice and sum
      # :pmf|:PMF : "peut mieux faire", reroll success and sum
      # :e2f|:E2F : "easy to fail", reroll success to confirm
      # :td       : "très difficile", only 6 are success
      # :tf       : "très facile", only 1 are failures
      #
      # @param [Array(Symbol)|Array(String)] list
      # @param [:pmf|:sc] priority if :sc and :pmf are available, choose only one
      def initialize(*list, priority: :pmf)
        @list = list.map { _1.is_a?(String) ? STR_TO_SYM[_1] : _1 }
        @priority = priority.is_a?(String) ? STR_TO_SYM[priority] : priority
        reduce!
      end

      %i[SC sc PMF pmf E2F e2f].each do |modifier|
        define_method(:"#{modifier}?") do
          @reduced == modifier
        end
      end

      def td?
        (@list.count(:td) - @list.count(:tf)) > 0
      end

      def tf?
        (@list.count(:tf) - @list.count(:td)) > 0
      end

      def to_h
        { list: @list, reduced: @reduced, priority: @priority }
      end

      # @param [:sc|:SC|:pmf|:PMF|:e2f] modifier. It will autoconvert strings.
      def <<(modifier)
        @list << modifier.is_a?(String) ? STR_TO_SYM[modifier] : modifier
        reduce!
        self
      end

      STR_TO_SYM = {
        "pmf" => :pmf,
        "PMF" => :PMF,
        "sc" => :sc,
        "SC" => :SC,
        "e2f" => :e2f,
        "E2F" => :E2F,
        "TD" => :td,
        "TF" => :tf,
        "td" => :td,
        "tf" => :tf,
      }.freeze

      PRIORITY_MAP = {
        big: {
          pmf: :PMF,
          sc: :SC,
        },
        small: {
          pmf: :pmf,
          sc: :sc,
        },
      }.freeze

      private def reduce!
        count_sc = @list.count(:sc) + (2 * @list.count(:SC))
        count_pmf = @list.count(:pmf) + (2 * @list.count(:PMF))
        count_e2f = @list.count(:e2f) + (2 * @list.count(:E2F))
        count_bonus = count_sc + count_pmf
        count_malus = count_e2f
        count_total = count_bonus - count_malus

        # puts "count_sc = #{count_sc}"
        # puts "count_pmf = #{count_pmf}"
        # puts "count_e2f = #{count_e2f}"
        # puts "count_bonus = #{count_bonus}"
        # puts "count_malus = #{count_malus}"
        # puts "count_total = #{count_total}"
        # TODO: use cond for pattern matching rather than if/else shit
        @reduced =
          if count_total == 0
            nil
          elsif count_total >= 2 && (count_sc >= 2 || count_pmf >= 2)
            PRIORITY_MAP[:big][
              if count_sc >= 2 && count_pmf >= 2
                @priority
              elsif count_sc >= 2
                :sc
              else
                :pmf
              end
            ]
          elsif count_total >= 2
            PRIORITY_MAP[:small][@priority]
          elsif count_total == 1
            PRIORITY_MAP[:small][
              if count_sc >= 1 && count_pmf >= 1
                @priority
              elsif count_sc >= 1
                :sc
              else
                :pmf
              end
            ]
          else # count_total < 0
            count_total == -1 ? :e2f : :E2F
          end
      end

      # Given the current modifiers, give the list of dice that are success
      def successes
        return 6..6 if td?
        return 2..6 if tf?

        4..6
      end

      # Given the current modifeirs, tell if a die is a success or not
      def success?(int)
        successes.include?(int)
      end

      DICE = (1..6).freeze

      def roll(n)
        rolled = Modifiers.raw_roll(n)
        success = rolled.count { success?(_1) }
        failures = n - success

        if self.PMF?
          rolled += Modifiers.raw_roll(success)
        elsif self.pmf?
          rolled += Modifiers.raw_roll(1) if success >= 1
        elsif self.SC?
          rolled += Modifiers.raw_roll(failures)
        elsif self.sc?
          rolled += Modifiers.raw_roll(1) if failures >= 1
        elsif self.E2F?
          rolled.delete_if { success?(_1) }
          rolled += Modifiers.raw_roll(success)
        elsif self.e2f?
          if success >= 1
            rolled.delete_at(-1)
            rolled += Modifiers.raw_roll(1)
          end
        end
        rolled
      end

      def self.raw_roll(n)
        Array.new(n) { Kernel.rand(DICE) }.sort
      end

      # def failures
      #   return [1, 2, 3] if @td && @td
      #   return [1, 2, 3, 4, 5] if @td
      #   return [1] if @tf

      #   [1, 2, 3]
      # end
    end
  end
end
